#!/usr/bin/perl
#
########################################################################################################################
########################################################################################################################
##
##      Copyright (C) 2019 Peter Walsh, Milford, NH 03055
##      All Rights Reserved
##
##  FILE
##
##      DevServer.pl
##
##  DESCRIPTION
##
##      Dev server startup
##
##  USAGE
##
##      perl DevServer [-v]
##
##      where:      -v      Verbose: print out things as they happen
##
########################################################################################################################
########################################################################################################################

use strict;
use warnings;
use Carp;

use English;
use POSIX ":sys_wait_h";
use Sys::Syslog qw(:standard :macros);

########################################################################################################################
########################################################################################################################
#
# User configuration variables
#

#
# End of user configuration
#
########################################################################################################################
#######################################################################################################################

########################################################################################################################
########################################################################################################################
##
## Data declarations
##
########################################################################################################################
########################################################################################################################

my $DevCmd   = "perl /home/Luminator/bin/Luminator.pl";

my %BackgroundCommands;

$SIG{ALRM} = \&Timeout;
$SIG{CHLD} = \&ChildExit;

#
# Colors for boot console messages
#
my $RED   = '\033[0;31m';
my $GREEN = '\033[0;32m';
my $NC    = '\033[0m';          # No Color

#
# Process command line args.
#
#     -v (verbose) means "print out extra stuff"
#
my $Verbose = 0;

$Verbose = 1
    if defined $ARGV[0] and $ARGV[0] eq "-v";

my $SysCommand;             # Currently executing system command (for debugging, in case timeout)
my $SysTimeout;             # Current timeout for executing system command

########################################################################################################################
########################################################################################################################
##
## Dev server
##
########################################################################################################################
########################################################################################################################

chdir("/home/Luminator/bin");

#
# Start the Dev server
#
# The -v option prints out debug information
#
Message("Luminator system is up and running.")
    if $Verbose;

BackgroundCommand($DevCmd);

while(1) {
    sleep(1);
    }

#######################################################################################################################
########################################################################################################################
##
## TimeoutCommand - Execute a system command, with timeout. Reboot if system times out
##
## Inputs:      System command to execute (ie: "ifconfig")
##              Timeout for command, in seconds (DEFAULT: 60)
##
## Outputs:     Text response of command, if successful
##
## NOTE: Reboots the system on timeout.
##
sub TimeoutCommand {
    $SysCommand = shift;            # Note: Make global for timeout/error function
    $SysTimeout = shift // 60;

    my $Rtnval;

    eval {
        alarm $SysTimeout;

        $Rtnval = `$SysCommand`;

        alarm 0;
        };

    ErrorReboot("Error executing $SysCommand ($!)")
        if $@;
        
    return $Rtnval;
    }


########################################################################################################################
########################################################################################################################
##
## Timeout - Process timeout of system command
##
## Inputs:      None (called as signal handler)
##
## Outputs:     None - reboots the system
##
sub Timeout { ErrorReboot("Timeout ($SysTimeout secs) executing $SysCommand ($!)"); }


########################################################################################################################
########################################################################################################################
##
## BackgroundCommand - Execute a system command in the background.
##
## Inputs:      System command to execute (ie: "ifconfig")
##
## Outputs:     None.
##
sub BackgroundCommand {
    my $Command = shift;
    my $PID;

    exec $Command
        unless $PID = fork();

    #
    # Make note of the PID and child command, in case it exits and we need to print out an error msg
    #
    $BackgroundCommands{$PID} = $Command;

    Message("BackgroundCommand $BackgroundCommands{$PID} (PID $PID)")
        if $Verbose;
    }


########################################################################################################################
########################################################################################################################
##
## EndBackgroundCommand - Kill a background command
##
## Inputs:      PID of background command already started
##
## Outputs:     None.
##
sub EndBackgroundCommand {
    my $PID = shift;

    Message("Stopping $BackgroundCommands{$PID} (PID $PID)")
        if $Verbose;

    #
    # Remove the $BackgroundCOmmands{$PID} first, so we know tha tthe child exit is expected.
    #
    delete $BackgroundCommands{$PID};

    kill "KILL",$PID;
    }


########################################################################################################################
########################################################################################################################
##
## ChildExit - Manage exit of background child process
##
## Inputs:      None (called as signal handler)
##
## Outputs:     None - reboots the system
##
sub ChildExit { 
    my $PID = waitpid(-1, WNOHANG);

    return
        if $PID == -1;      # == No child waiting

    return
        if $PID ==  0;      # == Children, none terminated

    my $Command = $BackgroundCommands{$PID};

    return
        unless defined $Command;

    #
    # If there's no entry in the BackgroundProcess list for thie PID, then it means the
    #   child was a direct command (and not meant to be in the background).
    #
    # Returning is expected, so ignore those.
    #
    print "SysCommand complete: $SysCommand\n"
        if $Verbose && !defined $Command && 
           substr($SysCommand,0,10) ne "sudo sh -c";   # Unnecessary, and messes up printed output

    return
        unless defined $Command;

    ErrorReboot("Reboot due to command exit: $Command (PID $PID)"); 
    }


########################################################################################################################
########################################################################################################################
##
## ConsoleMessage - Show message in boot screen
##
## Inputs:      Msg        Message to print
##              Fail       If message indicates a failure, (==1) print in Red, else (==0) print in Green
##
## Outputs:  None.
##
sub ConsoleMessage {
    my $Msg  = shift;
    my $Fail = shift // 0;

    return
        unless IAmRoot();

    if( $Fail ) { TimeoutCommand("sudo sh -c 'echo \"[${RED}FAILED${NC}] $Msg\"   >/dev/tty0'"); }
    else        { TimeoutCommand("sudo sh -c 'echo \"[${GREEN}  OK  ${NC}] $Msg\" >/dev/tty0'"); }
    }


########################################################################################################################
########################################################################################################################
##
## Message - Show message to the user
##
## Inputs:      Msg        Message to print
##              Fail       If message indicates a failure, (==1) print in Red, else (==0) print in Green
##
## Outputs:  None.
##
sub Message {
    my $Msg  = shift;
    my $Fail = shift // 0;

    #
    # Put the message in 3 places: System log, boot screen, and program log (captured from STDOUT)
    #
    ConsoleMessage ("**** DevServer: $Msg",$Fail);
    syslog(LOG_CRIT,"**** DevServer: $Msg");

    print "**** DevServer: $Msg\n";
    }


########################################################################################################################
########################################################################################################################
##
## ErrorReboot - Log an error, then reboot
##
## Inputs:      Message to log
##
## Outputs:     None - reboots the system
##
sub ErrorReboot {
    my $ErrorMsg = shift;

    Message("",1);
    Message("$ErrorMsg",1);

    if( NumUsers() > 0 ) {
        Message("No reboot, due to user login.",1);
        Message("",1);
        WatchUsers();
        }

    Message("Critical error - rebooting in 60 seconds.",1);
    Message("",1);
    Message("",1);

    sleep(60);

    if( NumUsers() > 0 ) {
        Message("No reboot, due to user login.",1);
        Message("",1);
        WatchUsers();
        }

    Message("Rebooting ",1);

    exec "sudo reboot";
    }


########################################################################################################################
########################################################################################################################
##
## WatchUsers - Watch for user logouts, then reboot the system
##
## Inputs:      None.
##
## Outputs:     None. Will reboot when users log out
##
## NOTE: If the user doesn't fix the problem but logs out, the system would never reboot. Keep watching
##         the user logins, and if they go to zero, reinstate the reboot timeout.
##
sub WatchUsers {

    while(1) {
        sleep 10;

        ErrorReboot("Reinstating reboot timer due to user logout.")
            unless NumUsers();
        }
    }


########################################################################################################################
########################################################################################################################
##
## NumUsers - Return number of logged-in users of the system
##
## Inputs:      None.
##
## Outputs:     Numer of logged in users
##
sub NumUsers { return scalar split /\s/, TimeoutCommand("users"); }


########################################################################################################################
########################################################################################################################
##
## IAmRoot - Return TRUE if caller is root
##
## Inputs:      None.
##
## Outputs:     TRUE  if caller is root
##              FALSE otherwise
##
sub IAmRoot {

    return $< == 0;
    }
