#!/usr/bin/perl -I .
#
########################################################################################################################
########################################################################################################################
##
##      Copyright (C) 2019 Peter Walsh, Milford, NH 03055
##      All Rights Reserved under the MIT license as outlined below.
##
##  FILE
##
##      Luminator.pl
##
##  DESCRIPTION
##
##      Run the Luminator sign
##
##      Read the bitmap files in /home/Luminator/Display and display them on the Luminator sign
##
##  USAGE
##
##      perl Luminator.pl [-v -b -t -d]
##
##          -v      Verbose: print out things as they happen
##
##          -b      Blank the LED display and exit
##
##          -t      Show the LED burnout test display and exit
##
##          -d      For debugging use
##
##  NOTES
##
##      The Luminator sign is an array of 115x14 LEDs which can be either ON or OFF.
##
##      Additionally, a flag in the header can set "BRIGHT" or "NORMAL" LED output. "NORMAL" is
##        quite readable for normal purposes.
##
##      The program will read and display bitmap files (single bit B/W palette) from the $Home/Images
##        based on the filenames as follows:
##
##          dd      Two digits - used for file ordering
##          "-"     A literal dash
##          dd      Two digits - the number of seconds to display the image (00 => infinite)
##          A       Optional "B" for "Bright" or "N" for normal.            (unspecified => normal)
##          *       Any further chars are ignored
##          ".bmp"  The image extension
##
##      The Luminator takes more than 1 second to change images. Waiting for less than 2 seconds
##        between images will not work.
##
##  EXAMPLE
##
##      Suppose the Images directory contains the following files:
##
##          01-04B-AmherstMakerspace.bmp
##          02-04N-AmherstMakerspace.bmp
##          03-02N-Blank.bmp
##
##      The first group (01 02 03) indicate the order in which the files are displayed.
##
##      The next groups indicate that the first file is displayed for 4 seconds as bright,
##        the second file is displayed for 4 seconds as normal, and the 3rd file is displayed
##        for 2 seconds in normal.
##
##      The extra text (AmherstMakerspace, Blank) is ignored, and can be used for documentation.
##
##      All other file types, including directories, are ignored.
##
########################################################################################################################
########################################################################################################################
##
##  MIT LICENSE
##
##  Permission is hereby granted, free of charge, to any person obtaining a copy of
##    this software and associated documentation files (the "Software"), to deal in
##    the Software without restriction, including without limitation the rights to
##    use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
##    of the Software, and to permit persons to whom the Software is furnished to do
##    so, subject to the following conditions:
##
##  The above copyright notice and this permission notice shall be included in
##    all copies or substantial portions of the Software.
##
##  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
##    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
##    PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
##    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
##    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
##    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##
########################################################################################################################
########################################################################################################################

use strict;
use warnings;
use Carp;

use Sys::Syslog  qw(:standard :macros);
use Getopt::Long qw(:config no_ignore_case);
use Image::BMP;
use File::Slurp;

our $VERSION = "2019.12_02";

########################################################################################################################
########################################################################################################################
##
## Data declarations
##
########################################################################################################################
########################################################################################################################

use constant LUMINATOR_DEVICE => "/dev/ttyUSB0";
use constant LUMINATOR_COLS   => 115;
use constant LUMINATOR_ROWS   => 14;

#my $DefaultDir = $ENV{"HOME"};
my $DefaultDir = "/home/Luminator";
my $DataDir    = "$DefaultDir/Display";

my %ImageTypes = map { $_ => 1 } ("bmp", "gif", "jpg", "png", "pnm", "ras", "tif");

my @ImageFiles;
my %Args;
my $Cache;

$| = 1;         # Flush output immediately

use constant START_FRAME    => "\xFC";
use constant END_FRAME      => "\xFB";
use constant ESC_CHAR       => "\x1B";
use constant BRIGHT_HEADER  => "\x00\x5F\x80\x01\x04\x11";
use constant NORMAL_HEADER  => "\x00\x50\x80\x01\x04\x11";

#
# EMERGENCY in bitmap form
#
my $TestMsg =
    "\x00\x00\x03\xf9\x86\x7f\x3f\x07\xe3\xf9\x86\x1e\x18\x60\x00\x00\x00".
    "\x00\x00\x03\xf9\x86\x7f\x3f\x87\xf3\xf9\xc6\x3f\x18\x60\x00\x00\x00".
    "\x00\x00\x03\x01\x86\x60\x31\xcc\x33\x01\xc6\x61\x98\x60\x00\x00\x00".
    "\x00\x00\x03\x01\xce\x60\x30\xcc\x03\x01\xc6\x61\x98\x60\x00\x00\x00".
    "\x00\x00\x03\x01\xce\x60\x30\xcc\x03\x01\xe6\x61\x98\x60\x00\x00\x00".
    "\x00\x00\x03\x01\xce\x60\x31\xcc\x03\x01\xe6\x60\x0c\xc0\x00\x00\x00".
    "\x00\x00\x03\xf1\xce\x7e\x3f\x8d\xf3\xf1\xa6\x60\x0f\xc0\x00\x00\x00".
    "\x00\x00\x03\xf1\xce\x7e\x3f\x0d\xf3\xf1\xb6\x60\x07\x80\x00\x00\x00".
    "\x00\x00\x03\x01\xfe\x60\x3c\x0c\x33\x01\xb6\x60\x03\x00\x00\x00\x00".
    "\x00\x00\x03\x01\xb6\x60\x36\x0c\x33\x01\x9e\x61\x83\x00\x00\x00\x00".
    "\x00\x00\x03\x01\xb6\x60\x33\x0c\x33\x01\x9e\x61\x83\x00\x00\x00\x00".
    "\x00\x00\x03\x01\x86\x60\x31\x8c\x33\x01\x9e\x61\x83\x00\x00\x00\x00".
    "\x00\x00\x03\xf9\x86\x7f\x30\xc7\xf3\xf9\x8e\x3f\x03\x00\x00\x00\x00".
    "\x00\x00\x03\xf9\x86\x7f\x30\x47\xe3\xf9\x8e\x1e\x03\x00\x00\x00\x00";

#
# Blank screen in bitmap form
#
my $BlankMsg = 
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00".
    "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00";


#
# All LEDs on - for burnout testing
#
my $BurnMsg = 
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF".
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF";

#
# Used in development - not really necessary
#
my $PMsg = [                                    ### Call Police ###
#            0xfc,
            0x00, 0x50, 0x80, 0x01, 0x01, 0x10,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x1f, 0x1f, 0x1f, 0x30, 0x01, 0xf0, 0x43, 0x06, 0x00, 0x07, 0xe3, 0xe6, 0x0f, 0x3e, 0x7e, 0xc0,
            0x3f, 0xbf, 0xbf, 0xb0, 0x03, 0xf8, 0xe3, 0x06, 0x00, 0x07, 0xf7, 0xf6, 0x0f, 0x7f, 0x7e, 0xc0,
            0x31, 0xb1, 0xb1, 0xb0, 0x03, 0x19, 0xb3, 0x06, 0x00, 0x06, 0x36, 0x36, 0x06, 0x63, 0x60, 0xc0,
            0x30, 0x31, 0xb0, 0x30, 0x03, 0x03, 0x1b, 0x00, 0x06, 0x00, 0x06, 0x36, 0x36, 0x06, 0x60, 0x60, 0xc0,
            0x3e, 0x31, 0xbe, 0x30, 0x03, 0x03, 0x1b, 0x00, 0x06, 0x00, 0x07, 0xf6, 0x36, 0x06, 0x60, 0x7c, 0xc0,
            0x1f, 0x31, 0x9f, 0x30, 0x03, 0x03, 0x1b, 0xe0, 0x06, 0x00, 0x07, 0xe6, 0x36, 0x06, 0x60, 0x7c, 0xc0,
            0x01, 0xb1, 0x81, 0x80, 0x03, 0x03, 0x1b, 0xe0, 0x06, 0x00, 0x06, 0x06, 0x36, 0x06, 0x60, 0x60, 0x00,
            0x31, 0xb1, 0xb1, 0x80, 0x03, 0x1b, 0x00, 0x1b, 0x00, 0x06, 0x00, 0x06, 0x06, 0x36, 0x06, 0x63, 0x60, 0x00,
            0x3f, 0xbf, 0xbf, 0xb0, 0x03, 0x1b, 0xe0, 0x1b, 0x00, 0xf7, 0xe0, 0x06, 0x07, 0xf7, 0xef, 0x7f, 0x7e, 0xc0,
            0x1f, 0x1f, 0x1f, 0x30, 0x01, 0xf3, 0x1b, 0x00, 0xf7, 0xe0, 0x06, 0x03, 0xe7, 0xef, 0x3e, 0x7e, 0xc0,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
#            0xc6,
#            0xfb
    ];

########################################################################################################################
########################################################################################################################
##
## Argument processing
##
########################################################################################################################
########################################################################################################################

use FindBin qw($Bin $RealScript);

exit HELP_MESSAGE()
    unless GetOptions(\%Args,
        "v", "t", "b", "d",
        "h|help", "V|version");

exit HELP_MESSAGE()
    if $Args{h};

exit VERSION_MESSAGE()
    if $Args{V};

sub HELP_MESSAGE {
    #
    # Print the text in the preamble of the script (at the top of this file) 
    #   for the help message.
    #
    my @Usage = read_file("$Bin/$RealScript");

    shift @Usage
        until $Usage[0] =~ /USAGE/;

    shift @Usage;           # Skip USAGE line

    print "\n";
    print "Usage: ";
    while(1) {
        my $Line = shift @Usage;
        last
            if $Line =~ /EXAMPLE/;

        $Line =~ s/\#\#//;

        print $Line;
        }

    return 0;
    }

sub VERSION_MESSAGE {
    print "\n";
    print basename($0) . " version $VERSION\n";
    print "\n";
    
    return 0;
    }

########################################################################################################################
########################################################################################################################
#
# Setup a simple log output to LUMINATOR.log in /var/syslog
#
openlog("Luminator", $Args{v} ? "perror" : "",LOG_LOCAL0);    # Print to stderr if verbose

my $Mask = LOG_MASK(LOG_NOTICE) | LOG_MASK(LOG_ERR) | LOG_MASK(LOG_CRIT);

$Mask |= LOG_MASK(LOG_DEBUG)
    if $Args{v};

setlogmask($Mask);

LogNotice("Luminator started");

print "\n"
    foreach 0 .. 50;

########################################################################################################################
########################################################################################################################
#
# In debug mode, loop display the known good EMERGENCY message
#
if( $Args{d} ) {

    LogNotice("Displaying debuggery");

    DisplayData($BlankMsg,0);
    sleep(2);

    while(1) {
        DisplayData($TestMsg,0);
        sleep(2);

        DisplayData($TestMsg,1);
        sleep(2);

        DisplayData($BlankMsg,0);
        sleep(2);
        }
    }


########################################################################################################################
########################################################################################################################
#
# In test mode, light up all the LEDs and exit
#
if( $Args{t} ) {

    LogNotice("Displaying LED burnout test");

    DisplayData($BurnMsg,0);
    sleep(2);

    exit(0);
    }


########################################################################################################################
########################################################################################################################
#
# In blank mode, turn off the LEDs and exit
#
if( $Args{b} ) {

    LogNotice("Turning off all LEDs");

    DisplayData($BlankMsg,0);
    sleep(2);

    exit(0);
    }


########################################################################################################################
########################################################################################################################
#
# Main loop
#
while(1) {
    chdir $DataDir;

    @ImageFiles = <"*.*">;

    DisplayFile($_)
        foreach @ImageFiles;
    }


########################################################################################################################
########################################################################################################################
#
# DisplayFile - Display image file on Luminator sign
#
# Inputs:      Filename of image to display
#
# Outputs:     None.
#
sub DisplayFile {
    my $ImageFile = shift;

    return
        unless -f -r $ImageFile;                        # Normal file (ie - not directory) and readable

    return
        unless $ImageTypes{substr($ImageFile,-3)};      # Not a type (by extension) we recognize

    my $ImageData = FindCache($ImageFile);

    LogNotice("Show $ImageFile (" . ($ImageData->{Bright} ? "Bright" : "Normal") . 
              ") for $ImageData->{Time} seconds...")
        if $Args{v};

    DisplayData($ImageData->{Pixels},$ImageData->{Bright});
    sleep($ImageData->{Time});
    }


########################################################################################################################
########################################################################################################################
#
# FindCache - Add file contents to our internal cache
#
# Inputs:      Filename of image to find
#
# Outputs:     Cached data for image file
#
sub FindCache {
    my $ImageFile = shift;

    #
    # Use existing data if the file hasn't changed.
    #
    my $ImageFileLastMod = (stat($ImageFile))[9];

    return $Cache->{$ImageFile}
        if defined $Cache->{$ImageFile} and
           $Cache->{$ImageFile}{LastMod} == $ImageFileLastMod;

    ####################################################################################################################
    #
    # Otherwise, convert the file and read the data to cache
    #
    my ($FileNo, $FileTime, $Bright, $Other, $Ext) = $ImageFile =~ m/(\d+)-(\d+)([BbNn])(.*)\.(...)$/;

    $Cache->{$ImageFile}{LastMod} = $ImageFileLastMod;
    $Cache->{$ImageFile}{   Time} = 0+$FileTime;
    $Cache->{$ImageFile}{ Bright} = ((uc $Bright) // "N") eq "B";

    `mogrify -format pbm -path /tmp -gravity NorthWest -extent 136x14 $ImageFile`;

    my $PBMFile = $ImageFile;

    substr($PBMFile,-3) = "pbm";      # Change to new file extension in our local name var

    my @Lines = read_file("/tmp/$PBMFile");
    $Cache->{$ImageFile}{Pixels} = LuminatorEncode($Lines[2]);

    #
    # Time == 0 means forever.
    #
    # Also, enforce the "minimum 2-second" rule
    #
    $Cache->{$ImageFile}{Time} = 999999999
        if $Cache->{$ImageFile}{Time} == 0;

    $Cache->{$ImageFile}{Time} = 2
        if $Cache->{$ImageFile}{Time} < 2;

    #
    # Return the cached data
    #
    return $Cache->{$ImageFile};
    }


########################################################################################################################
########################################################################################################################
#
# DisplayData - Display image data on luminator
#
# Inputs:      Data to display
#              TRUE  if output should be BRIGHT
#
# Outputs:     None.
#
sub DisplayData {
    my $ImageData = shift;
    my $Bright    = shift // 0;

    my $SignData;

    if( $Bright ) { $SignData .= BRIGHT_HEADER; }
    else          { $SignData .= NORMAL_HEADER; }

    $SignData .= $ImageData;
    $SignData .= Checksum($SignData);

    write_file(LUMINATOR_DEVICE, {binmode => ":raw"},START_FRAME . $SignData . END_FRAME);
    }


########################################################################################################################
########################################################################################################################
#
# Checksum - Return the checksum of image data
#
# Inputs:      Data to checksum
#
# Outputs:     (char) checksum of data
#
sub Checksum {
    my $ImageData = shift;

    my $Checksum  = 0;

    $Checksum += ord(substr($ImageData,$_,1))
        foreach 0 .. length($ImageData)-1;

    return chr($Checksum & 0xFF);
    }


########################################################################################################################
########################################################################################################################
#
# LuminatorEncode - Change the data so that ESC_CHAR, START_FRAME, and END_FRAME become the escaped versions
#
# Inputs:      Data to change
#
# Outputs:     New data, appropriately formatted
#
sub LuminatorEncode {
    my $ImageData = shift;

    my $LuminatorData = "";

    for( my $i = 0; $i < length($ImageData); $i++ ) {
        my $ImageChar = substr($ImageData,$i,1);

        #
        # The escape char itself is encoded as (0x1B,0x00)
        #
        if(   $ImageChar eq ESC_CHAR ) {
            $LuminatorData .= ESC_CHAR;
            $ImageChar    = chr(0x00);
            }
        #
        # START_FRAME is escaped, lest we confuse the display with a 2nd start frame
        #
        elsif( $ImageChar eq START_FRAME ) {
            $LuminatorData .= ESC_CHAR;
            $ImageChar    = chr(0xE1);
            }
        #
        # END_FRAME is escaped, lest we confuse the display with a premature ending
        #
        elsif( $ImageChar eq END_FRAME ) {
            $LuminatorData .= ESC_CHAR;
            $ImageChar    = chr(0xE0);
            }

        $LuminatorData .= $ImageChar;
        }

    return $LuminatorData;
    }


########################################################################################################################
########################################################################################################################
#
# LogNotice - Log informational messages
#
# Inputs:      Message to log
#
# Outputs:     undef
#
# NOTE: Used for non-error messages, does not set the prev Error 
#
sub LogNotice {
    my $Msg = shift;

    syslog(LOG_NOTICE,$Msg);

    return undef;
    }
